/*
************************************************************************
*
*   %M% - Windows/OpenGL image dump
*
*   Copyright (c) 1996
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : %E%
*   Pathname of SCCS file     : %P%
*   SCCS identification       : %I%
*
************************************************************************
*/

#include "winogl.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>

#include <tiffio.h>

#include <sg.h>
#include <sg_get.h>
#include <sg_attr.h>
#include <io_error.h>

static FILE *TiffFileP;
static SgPlotOrient Orient;
static int ImgW, ImgH;
static float Gamma;
static float OldVpX, OldVpY, OldVpW, OldVpH;

IORes
IOWinOGLStartDump(char *fileName,
    SgPlotOrient orient, int w, int h, float gamma)
{
  int i;

  TiffFileP = fopen(fileName, "w");
  if (TiffFileP == NULL) {
    IORaiseError("could not open output file");
    return IO_RES_ERR;
  }

  Orient = orient;
  ImgW = w;
  ImgH = h;
  Gamma = gamma;

  SgSetAttr();
  SgUpdateView();

  SgGetViewport(&OldVpX, &OldVpY, &OldVpW, &OldVpH);
  SgSetViewport(0.0, 0.0, w, h);

  for (i = 0; i < SG_FEATURE_NO; i++)
    SgSetFeature(i, SgGetFeature(i));

  return IO_RES_OK;
}

static IORes
writeTiff(void)
{
  TIFF *tif;
  int tiffW, tiffH;
  int rowsPerStrip, bufSize, rowI, i;
  float *fBuf;
  unsigned char *cBuf;
  IORes res;

  tif = TIFFFdOpen(fileno(TiffFileP), "output file", "w");
  if (tif == NULL) {
    IORaiseError("could not create TIFF file");
    return IO_RES_ERR;
  }

  if (Orient == SG_PLOT_NORMAL) {
    tiffW = ImgW;
    tiffH = ImgH;
    bufSize = 4 * ((3 * tiffW + 3) / 4);
    glPixelStorei(GL_PACK_ALIGNMENT, 4);
  } else {
    tiffW = ImgH;
    tiffH = ImgW;
    bufSize = 3 * tiffW;
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
  }

  rowsPerStrip = (8 * 1024) / (3 * tiffW);
  if (rowsPerStrip == 0)
    rowsPerStrip = 1;

  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, tiffW);
  TIFFSetField(tif, TIFFTAG_IMAGELENGTH, tiffH);
  TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(tif, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
  TIFFSetField(tif, TIFFTAG_DOCUMENTNAME, "MOLMOL Plot");
  TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "MOLMOL Plot");
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 3);
  TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);

  cBuf = malloc(bufSize * sizeof(*cBuf));
  if (Gamma != 1.0)
    fBuf = malloc(bufSize * sizeof(*fBuf));

  res = IO_RES_OK;
  for (rowI = 0; rowI < tiffH; rowI++) {
    if (Gamma == 1.0) {
      if (Orient == SG_PLOT_NORMAL)
	glReadPixels(0, ImgH - 1 - rowI, ImgW, 1,
	    GL_RGB, GL_UNSIGNED_BYTE, cBuf);
      else
	glReadPixels(rowI, 0, 1, ImgH,
	    GL_RGB, GL_UNSIGNED_BYTE, cBuf);
    } else {
      if (Orient == SG_PLOT_NORMAL)
	glReadPixels(0, ImgH - 1 - rowI, ImgW, 1,
	    GL_RGB, GL_FLOAT, fBuf);
      else
	glReadPixels(rowI, 0, 1, ImgH,
	    GL_RGB, GL_FLOAT, fBuf);
      
      for (i = 0; i < 3 * tiffW; i++)
	cBuf[i] = (int) (255.99 * powf(fBuf[i], Gamma));
    }

    if (TIFFWriteScanline(tif, cBuf, rowI, 0) < 0) {
      IORaiseError("error while writing TIFF file");
      res = IO_RES_ERR;
      break;
    }
  }

  free(cBuf);
  if (Gamma != 1.0)
    free(fBuf);

  TIFFFlushData(tif);
  TIFFClose(tif);

  return res;
}

IORes
IOWinOGLEndDump(void)
{
  IORes res;

  res = writeTiff();

  (void) fclose(TiffFileP);

  SgSetAttr();
  SgSetViewport(OldVpX, OldVpY, OldVpW, OldVpH);

  return res;
}
